﻿/*
  DXL APIを使う
   複数軸への角速度指令
*/

#include  <stdio.h>
#include  <pthread.h>
#include  "dx2lib.h"

#define COMPORT   "/dev/ttyUSB0"  // ポート名
#define BAUDRATE  (57600)         // ボーレート[bps]

//-----------------------------------------------------------
// 予約語
//-----------------------------------------------------------
// スレッド伝達用構造体
typedef struct {
  TDeviceID dev;      // デバイスID
  const uint8_t *ids; // IDの配列
  int num;            // ID数
} TMonInfo;

//-----------------------------------------------------------
// グローバル変数
//-----------------------------------------------------------
bool term = false;  // スレッド終了フラグ

//===========================================================
// ミリ秒スリープ
//===========================================================
void Sleep (int milliseconds) {
  struct timespec ts;
  ts.tv_sec = milliseconds / 1000;
  ts.tv_nsec = (milliseconds % 1000) * 1000000;
  nanosleep(&ts, NULL);
}

//===========================================================
// 複数軸の現在角度・角速度・エラー情報をモニタするスレッド
//===========================================================
void *monthread (void *pdev) {
  TMonInfo *info = (TMonInfo *)pdev;
  double  pangle[info->num], pvelo[info->num], pcur[info->num];
  for (int i = 0; i < info->num; i++) pangle[i] = pvelo[i] = pcur[i] = 0.0;
  while (!term) {
    // 現在位置取得
    DXL_GetPresentAngles (info->dev, info->ids, pangle, info->num);
    // 現在角速度取得
    DXL_GetPresentVelocities (info->dev, info->ids, pvelo, info->num);
    // 現在電流取得
    DXL_GetPresentCurrents(info->dev, info->ids, pcur, info->num);
    for (int i = 0; i < info->num; i++)
      printf ("(%d:$%04X %4.0f%6.1f)", info->ids[i], DXL_GetErrorCode (info->dev, info->ids[i]), pangle[i], pvelo[i]);
    printf("\r");
    Sleep (10);
  }
  return NULL;
}

//===========================================================
// メイン
//===========================================================
int main (void) {
  TMonInfo info;
  pthread_t th;

  // 検出したDynamixelのID一覧
  uint8_t ids[253];
  // 検出したDynamixel数
  int num = 0;

  TDeviceID dev = DX2_OpenPort (COMPORT, BAUDRATE);
  if (dev != 0) {
    printf ("Open success\n");

    // 接続された全デバイスを検索しその一覧を表示
    printf ("scanning devices...");
    num = DXL_ScanDevices (dev, ids);
    printf (" %d device detected.\n", num);
    DXL_PrintDevicesList ((void *)&printf);

    // モニタスレッド開始
    info.dev = dev;
    info.ids = ids;
    info.num = num;
    if (pthread_create (&th, NULL, monthread, (void *)&info)) {
    } else {
      // 速度指令用変数
      double velos[num];
      // Velocityモードに設定
      if (DXL_SetOperatingModesEquival (dev, ids, num, 1)) {
        // 制御開始
        DXL_SetTorqueEnablesEquival (dev, ids, num, true);

        for (int i = 0; i < num; i++) velos[i] = 30;  // 30deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (2000);
        for (int i = 0; i < num; i++) velos[i] = 50;  // 50deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (2000);
        for (int i = 0; i < num; i++) velos[i] = 90;  // 90deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (2000);
        for (int i = 0; i < num; i++) velos[i] = -90; // -90deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (2000);
        for (int i = 0; i < num; i++) velos[i] = -50; // -50deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (2000);
        for (int i = 0; i < num; i++) velos[i] = -30; // -30deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (2000);
        for (int i = 0; i < num; i++) velos[i] = 0;   // 0deg/s
        DXL_SetGoalVelocities (dev, ids, velos, num);
        Sleep (500);

        // 制御停止
        DXL_SetTorqueEnablesEquival (dev, ids, num, false);
      }
      // モニタスレッド終了
      term = true;
      pthread_join (th, NULL);
    }

    DX2_ClosePort (dev);
  }
  printf("\n");
}
